@@ -4,7 +4,7 @@
 
     Flask-Security Unified Signin module
 
-    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2024 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
     This implements a unified sign in endpoint - allowing
@@ -29,10 +29,12 @@
 
 """
 
+from __future__ import annotations
+
 import time
 import typing as t
 
-from flask import current_app as app
+from flask import current_app
 from flask import after_this_request, request, session
 from flask_login import current_user
 from wtforms import (
@@ -50,6 +52,7 @@ from wtforms.widgets import CheckboxInput
 from .confirmable import requires_confirmation
 from .decorators import anonymous_user_required, auth_required, unauth_csrf
 from .forms import (
+    _setup_methods_xlate,
     Form,
     NextFormMixin,
     Required,
@@ -76,6 +79,8 @@ from .utils import (
     get_message,
     get_url,
     get_within_delta,
+    is_user_authenticated,
+    localize_callback,
     json_error_response,
     login_user,
     lookup_identity,
@@ -84,6 +89,7 @@ from .utils import (
     url_for_security,
     view_commit,
 )
+from .twofactor import tf_clean_session
 from .webauthn import has_webauthn
 
 if t.TYPE_CHECKING:  # pragma: no cover
@@ -141,7 +147,7 @@ class _UnifiedPassCodeForm(Form):
     """Common form for signin and verify/reauthenticate."""
 
     # filled in by caller
-    user: "User"
+    user: User
 
     # Filled in here
     authn_via: str
@@ -358,7 +364,7 @@ class UnifiedSigninSetupValidateForm(Form):
     """The unified sign in setup validation form"""
 
     # These 2 filled in by view
-    user: "User"
+    user: User
     totp_secret: str
 
     passcode = StringField(
@@ -407,8 +413,8 @@ def _send_code_helper(form, send_magic_link):
 
 
 @anonymous_user_required
-@unauth_csrf(fall_through=True)
-def us_signin_send_code() -> "ResponseValue":
+@unauth_csrf()
+def us_signin_send_code() -> ResponseValue:
     """
     Send code view. POST only.
     This takes an identity (as configured in USER_IDENTITY_ATTRIBUTES)
@@ -445,7 +451,7 @@ def us_signin_send_code() -> "ResponseValue":
         )
     elif request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
         # TODO - this suppresses the error if they don't select ANY send code option.
-        rinfo: t.Dict[str, t.Dict[str, str]] = dict(
+        rinfo: dict[str, dict[str, str]] = dict(
             identity=dict(), passcode=dict(), chosen_method=dict()
         )
         form_errors_munge(form, rinfo)
@@ -481,7 +487,7 @@ def us_signin_send_code() -> "ResponseValue":
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def us_verify_send_code() -> "ResponseValue":
+def us_verify_send_code() -> ResponseValue:
     """
     Send code during verify. POST only.
     """
@@ -527,8 +533,8 @@ def us_verify_send_code() -> "ResponseValue":
     )
 
 
-@unauth_csrf(fall_through=True)
-def us_signin() -> "ResponseValue":
+@unauth_csrf()
+def us_signin() -> ResponseValue:
     """
     Unified sign in view.
     This takes an identity (as configured in USER_IDENTITY_ATTRIBUTES)
@@ -540,7 +546,7 @@ def us_signin() -> "ResponseValue":
     For POST - redirects to POST_LOGIN_VIEW (forms) or returns 400 (json).
     """
 
-    if current_user.is_authenticated and request.method == "POST":
+    if is_user_authenticated(current_user) and request.method == "POST":
         # Just redirect current_user to POST_LOGIN_VIEW (or next).
         # While its tempting to try to logout the current user and login the
         # new requested user - that simply doesn't work with CSRF.
@@ -560,6 +566,9 @@ def us_signin() -> "ResponseValue":
     form = t.cast(UnifiedSigninForm, build_form_from_request("us_signin_form"))
     form.submit.data = True
     form.submit_send_code.data = False
+    # Clean out any potential old session info - in case of previous
+    # aborted 2FA attempt.
+    tf_clean_session()
 
     if form.validate_on_submit():
         # Check if multi-factor is required. Some (this is configurable) don't
@@ -598,7 +607,7 @@ def us_signin() -> "ResponseValue":
         }
         return base_render_json(form, include_user=False, additional=payload)
 
-    if current_user.is_authenticated:
+    if is_user_authenticated(current_user):
         # Basically a no-op if authenticated - just perform the same
         # post-login redirect as if user just logged in.
         return redirect(get_post_login_redirect())
@@ -625,7 +634,7 @@ def us_signin() -> "ResponseValue":
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def us_verify() -> "ResponseValue":
+def us_verify() -> ResponseValue:
     """
     Re-authenticate to reset freshness time.
     This is likely the result of a reauthn_handler redirect, which
@@ -672,28 +681,28 @@ def us_verify() -> "ResponseValue":
 
 
 @anonymous_user_required
-def us_verify_link() -> "ResponseValue":
+def us_verify_link() -> ResponseValue:
     """
     Used to verify a magic email link. GET only
     Since this is just a URL - be careful not to disclose info like
     whether email exists or not.
     """
-    email = request.args.get("email", None)
+    fs_uniquifier = request.args.get("id", None)
     code = request.args.get("code", None)
-    if not email or not code:
+    if not fs_uniquifier or not code:
         m, c = get_message("API_ERROR")
-        if _security.redirect_behavior == "spa":
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(get_url(cv("LOGIN_ERROR_VIEW"), qparams={c: m}))
         do_flash(m, c)
         return redirect(url_for_security("us_signin"))
 
-    user = _datastore.find_user(email=email)
+    user = _datastore.find_user(fs_uniquifier=fs_uniquifier)
     if not user or not user.active:
         if not user:
             m, c = generic_message("USER_DOES_NOT_EXIST", "GENERIC_AUTHN_FAILED")
         else:
             m, c = generic_message("DISABLED_ACCOUNT", "GENERIC_AUTHN_FAILED")
-        if _security.redirect_behavior == "spa":
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(get_url(cv("LOGIN_ERROR_VIEW"), qparams={c: m}))
         do_flash(m, c)
         return redirect(url_for_security("us_signin"))
@@ -706,7 +715,7 @@ def us_verify_link() -> "ResponseValue":
         window=cv("US_TOKEN_VALIDITY"),
     ):
         m, c = generic_message("INVALID_CODE", "GENERIC_AUTHN_FAILED")
-        if _security.redirect_behavior == "spa":
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(
                 get_url(
                     cv("LOGIN_ERROR_VIEW"),
@@ -728,7 +737,7 @@ def us_verify_link() -> "ResponseValue":
         # isn't quite ready for SPA. So we return an error via a redirect rather
         # than mess up SPA applications. To be clear - this simply doesn't
         # work - using a magic link w/ 2FA - need to use code.
-        if _security.redirect_behavior == "spa":
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(
                 get_url(
                     cv("LOGIN_ERROR_VIEW"),
@@ -743,7 +752,7 @@ def us_verify_link() -> "ResponseValue":
 
     login_user(user, authn_via=["email"])
     after_this_request(view_commit)
-    if _security.redirect_behavior == "spa":
+    if cv("REDIRECT_BEHAVIOR") == "spa":
         # We do NOT send the authentication token here since the only way to
         # send it would be via a query param and that isn't secure. (logging and
         # possibly HTTP Referer header).
@@ -762,7 +771,7 @@ def us_verify_link() -> "ResponseValue":
     within=lambda: cv("FRESHNESS"),
     grace=lambda: cv("FRESHNESS_GRACE_PERIOD"),
 )
-def us_setup() -> "ResponseValue":
+def us_setup() -> ResponseValue:
     """
     Change unified sign in methods.
     We want to verify the new method - so don't store anything yet in DB
@@ -780,6 +789,16 @@ def us_setup() -> "ResponseValue":
         c for c in form.delete_choices if c[0] in active_methods
     ]
 
+    # translate active methods
+    if not active_methods:
+        active_methods = [None]
+    current_methods = _security.i18n_domain.format_list(
+        [localize_callback(_setup_methods_xlate[m]) for m in active_methods]
+    )
+    current_methods_msg = get_message(
+        "US_CURRENT_METHODS", method_list=current_methods
+    )[0]
+
     if form.validate_on_submit():
         qrcode_values = dict()
         json_response = dict()
@@ -792,6 +811,15 @@ def us_setup() -> "ResponseValue":
             for m in delete_method:
                 _datastore.us_reset(current_user, m)
             active_methods = _compute_active_methods(current_user)
+            if not active_methods:
+                active_methods = [None]
+            current_methods = _security.i18n_domain.format_list(
+                [localize_callback(_setup_methods_xlate[m]) for m in active_methods]
+            )
+
+            current_methods_msg = get_message(
+                "US_CURRENT_METHODS", method_list=current_methods
+            )[0]
             form.chosen_method.choices = [
                 c for c in form.setup_choices if c[0] not in active_methods
             ]
@@ -800,7 +828,8 @@ def us_setup() -> "ResponseValue":
             ]
             form.delete_method.data = None
             us_profile_changed.send(
-                app._get_current_object(),  # type: ignore
+                current_app._get_current_object(),  # type: ignore
+                _async_wrapper=current_app.ensure_sync,
                 user=current_user,
                 methods=delete_method,
                 delete=True,
@@ -841,6 +870,7 @@ def us_setup() -> "ResponseValue":
                     cv("US_SETUP_TEMPLATE"),
                     available_methods=cv("US_ENABLED_METHODS"),
                     active_methods=active_methods,
+                    current_methods_msg=current_methods_msg,
                     setup_methods=setup_methods,
                     us_setup_form=form,
                     **_security._run_ctx_processor("us_setup"),
@@ -877,6 +907,7 @@ def us_setup() -> "ResponseValue":
             cv("US_SETUP_TEMPLATE"),
             available_methods=cv("US_ENABLED_METHODS"),
             active_methods=active_methods,
+            current_methods_msg=current_methods_msg,
             setup_methods=setup_methods,
             code_sent=form.chosen_method.data in _compute_code_methods(),
             chosen_method=form.chosen_method.data,
@@ -907,6 +938,7 @@ def us_setup() -> "ResponseValue":
         cv("US_SETUP_TEMPLATE"),
         available_methods=cv("US_ENABLED_METHODS"),
         active_methods=active_methods,
+        current_methods_msg=current_methods_msg,
         setup_methods=setup_methods,
         us_setup_form=form,
         **_security._run_ctx_processor("us_setup"),
@@ -914,7 +946,7 @@ def us_setup() -> "ResponseValue":
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def us_setup_validate(token: str) -> "ResponseValue":
+def us_setup_validate(token: str) -> ResponseValue:
     """
     Validate new setup.
     The token is the state variable which is signed and timed
@@ -949,7 +981,8 @@ def us_setup_validate(token: str) -> "ResponseValue":
         _datastore.us_set(current_user, method, state["totp_secret"], phone)
 
         us_profile_changed.send(
-            app._get_current_object(),  # type: ignore
+            current_app._get_current_object(),  # type: ignore
+            _async_wrapper=current_app.ensure_sync,
             user=current_user,
             methods=[method],
             delete=False,
@@ -994,13 +1027,13 @@ def us_send_security_token(
 
     .. versionadded:: 3.4.0
     """
-    token = _security._totp_factory.generate_totp_password(totp_secret)
+    code = _security._totp_factory.generate_totp_password(totp_secret)
 
     if method == "email":
         login_link = None
         if send_magic_link:
             login_link = url_for_security(
-                "us_verify_link", email=user.email, code=token, _external=True
+                "us_verify_link", id=str(user.fs_uniquifier), code=code, _external=True
             )
         send_mail(
             cv("US_EMAIL_SUBJECT"),
@@ -1008,12 +1041,12 @@ def us_send_security_token(
             "us_instructions",
             user=user,
             username=user.calc_username(),
-            token=token,  # deprecated
-            login_token=token,
+            token=code,  # deprecated
+            login_token=code,
             login_link=login_link,
         )
     elif method == "sms":
-        m, c = get_message("USE_CODE", code=token)
+        m, c = get_message("USE_CODE", code=code)
         from_number = cv("SMS_SERVICE_CONFIG")["PHONE_NUMBER"]
         to_number = phone_number
         sms_sender = SmsSenderFactory.createSender(cv("SMS_SERVICE"))
@@ -1023,13 +1056,14 @@ def us_send_security_token(
         # tokens are generated automatically with authenticator apps
         # and passwords are well passwords
         # Still go ahead and notify signal receivers that they requested it.
-        token = None
+        code = None
     us_security_token_sent.send(
-        app._get_current_object(),
+        current_app._get_current_object(),
+        _async_wrapper=current_app.ensure_sync,
         user=user,
         method=method,
-        token=token,
-        login_token=token,
+        token=code,
+        login_token=code,
         phone_number=phone_number,
         send_magic_link=send_magic_link,
     )
